/**
 * Copyright (c) Members of the EMI Collaboration. 2011
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2011-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *
 */


/*!
    \page lcmaps_tracking_group_id.mod Tracking Group ID for batch systems acquiring plugin

    \section trackinggroupidsynopsis SYNOPSIS
    \b lcmaps_tracking_group_id.mod [--auto-discover] [--tracking-groupid-min \<number of secondary gid\> --tracking-groupid-max \<number of secondary gid\>]
 
    \section trackinggroupiddesc DESCRIPTION

The Tracking Group ID plugin will add one or more secondary Group IDs to the final mapping result. This is too ensure that the batch system's specially issues secondary GIDs are kept, even if a process like gLExec is switching the Unix/POSIX process ownership. This feature is required to ensure that stray processes from users are cleaned up from a cluster's Worker Node when the job is done.

The plugin can preserve a fixed range of Unix Group IDs or the plugin is able to auto-discover the tracking GIDs. The automatic discovery should be carefully used. It could for instance preserve AFS specific secondary Group ID. The auto discovery mechanism will lookup all secondary Group IDs of the running process and will preserve the Group IDs which are left nameless. This might come in handy, but this doesn't work if system administrators (for good reasons) assign group names to the tracking Group IDs.

    \section trackinggroupidoptions OPTIONS
    \subsection trackinggroupidoption1 --auto-discover
        The plugin will try to discover the Tracking GIDs assigned to the current process. This feature is not reliable due to it's direct tie to how system administrators can configure their cluster.

    \subsection trackinggroupidoption2 --tracking-groupid-min \<number of secondary gid\>
        Declare the lowest numerical representation of a secondary Group ID used as a Tracking Group ID. Note: the number specified is included in the Group ID match with an equal or higher statement.

    \subsection trackinggroupidoption3 --tracking-groupid-max \<number of secondary gid\>
        Declare the highest numerical representation of a secondary Group ID used as a Tracking Group ID. Note: the number specified is included in the Group ID match with an equal or lower statement.


    \section trackinggroupidreturnvalue RETURN VALUES
        \li LCMAPS_MOD_SUCCESS : Success
        \li LCMAPS_MOD_FAIL    : Failure

    \section trackinggroupiderrors ERRORS
        Email to grid-mw-security@nikhef.nl for help, or file a ticket in the most current bug tracker that we need to use.
 
    \section trackinggroupidSeeAlso SEE ALSO
        \ref lcmaps_posix_enf.mod "lcmaps_posix_enf.mod",
*/

/*!
    \file   lcmaps_tracking_groupid.c
    \brief  Tracking Group ID plugin
    \author Oscar Koeroo for the EMI project

    This file contains the code for the tracking groupid acquire LCMAPS plugin
    -# plugin_initialize()
    -# plugin_run()
    -# plugin_terminate()
    -# plugin_introspect()
*/

/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_tracking_groupid_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <grp.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <limits.h>

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
/******************************************************************************
                                Definitions
******************************************************************************/
#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1

#define MAX_UNDEFINED ((int)(-1))


/******************************************************************************
                          Module specific prototypes
******************************************************************************/

static int tracking_groupid_get_gidlist();
static int plugin_run_or_verify(int, lcmaps_argument_t *, int);


/******************************************************************************
                       Define module specific variables
******************************************************************************/

static int   set_auto = 0;
static gid_t set_groupid_min = 0;
static gid_t set_groupid_max = 0;



/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(
        int argc,
        char ** argv
)
{
    int i, j;
    char * overflow = NULL;
    char * logstr = "lcmaps_plugin_tracking_groupid-plugin_initialize()";

    lcmaps_log_debug(5,"%s: passed arguments:\n", logstr);
    for (i=0; i < argc; i++)
    {
       lcmaps_log_debug(5,"%s: arg %d is %s\n", logstr, i, argv[i]);
    }

    /*
     * Parse arguments, argv[0] = name of plugin, so start with i = 1
     */
    for (i = 1; i < argc; i++)
    {
        if ((strcmp(argv[i], "--auto-discover") == 0) || (strcmp(argv[i], "--autodiscover") == 0)){
            set_auto = 1;
        }
        else if (strcmp(argv[i], "--tracking-groupid-min") == 0){
            overflow = NULL;
            set_groupid_min = strtol(argv[i + 1], &overflow, 10);
            if ((set_groupid_min == 0) && (errno != 0)) {
                lcmaps_log(LOG_ERR,"%s: Error expected the Minimum Unix Group ID as a number not as a string. Got this \"%s\".\n", logstr, argv[i+1]);
                goto fail_tracking_groupid;
            }
            else if ((overflow != NULL) && (strlen(overflow) > 0)) {
                lcmaps_log(LOG_ERR,"%s: Error expected the Minimum Unix Group ID as a number not as a string. Got this \"%s\" and failed here \"%s\".\n", logstr, argv[i+1], overflow);
                goto fail_tracking_groupid;
            }
            i++;
        }
        else if (strcmp(argv[i], "--tracking-groupid-max") == 0){
            overflow = NULL;
            set_groupid_max = strtol(argv[i + 1], &overflow, 10);
            if ((set_groupid_max == 0) && (errno != 0)) {
                lcmaps_log(LOG_ERR,"%s: Error expected the Max Unix Group ID as a number not as a string. Got this \"%s\".\n", logstr, argv[i+1]);
                goto fail_tracking_groupid;
            }
            else if ((overflow != NULL) && (strlen(overflow) > 0)) {
                lcmaps_log(LOG_ERR,"%s: Error expected the Max Unix Group ID as a number not as a string. Got this \"%s\" and failed here \"%s\".\n", logstr, argv[i+1], overflow);
                goto fail_tracking_groupid;
            }
            i++;
        }
        else
        {
            lcmaps_log(LOG_ERR,"%s: Error in initialization parameter: %s (failure)\n", logstr, argv[i]);
            goto fail_tracking_groupid;
        }
    }

    if (set_auto && (set_groupid_min || set_groupid_max)){
        lcmaps_log(LOG_ERR, "%s: Misconfiguarion in the LCMAPS Tracking GroupID plug-in. You can't configure a min/max tracking GID and the auto discovery of tracking Group IDs.\n", logstr);
        goto fail_tracking_groupid;
    }

    /* TODO */
    if (set_auto) {
        lcmaps_log(LOG_ERR, "%s: Auto discovery of Tracking Group IDs is not yet implemented. Please send an email to: grid-mw-security@nikhef.nl\n", logstr);
        goto fail_tracking_groupid;
    }

    if (!set_groupid_min) {
        lcmaps_log(LOG_ERR, "%s: No minimum Unix Group ID configured.\n", logstr);
        goto fail_tracking_groupid;
    }
    if (!set_groupid_max) {
        lcmaps_log(LOG_ERR, "%s: No maximum Unix Group ID configured.\n", logstr);
        goto fail_tracking_groupid;
    }

    return LCMAPS_MOD_SUCCESS;

fail_tracking_groupid:
    return LCMAPS_MOD_FAIL;
}

/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(
        int * argc,
        lcmaps_argument_t ** argv
)
{
    char * logstr = "lcmaps_plugin_posix_enf-plugin_introspect()";  

    static lcmaps_argument_t argList[] = {
        {NULL           ,       NULL            , -1,   NULL}
    };

    lcmaps_log_debug(5,"%s: introspecting\n", logstr);

    *argv = argList;
    *argc = lcmaps_cntArgs(argList);
    lcmaps_log_debug(5,"%s: address first argument: 0x%x\n", logstr, argList);

    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(
        int argc,
        lcmaps_argument_t * argv
)

{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}

/******************************************************************************
+Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(
        int argc,
        lcmaps_argument_t * argv
)
{
    lcmaps_log(1, "The %s interface is not supported.\n", __func__);
    return LCMAPS_MOD_FAIL;
}

static int plugin_run_or_verify(
        int argc,
        lcmaps_argument_t * argv,
        int lcmaps_mode
)
{
    char * logstr = "lcmaps_plugin_tracking_groupid-plugin_plugin_run_or_verify()";  
    int   i = 0;
    int   n_sec_groups_list = 0;
    gid_t * sec_groups_list = NULL;

    if (lcmaps_mode == PLUGIN_RUN)
        logstr = "lcmaps_plugin_tracking_groupid-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = "lcmaps_plugin_tracking_groupid-plugin_verify()";
    else
    {
        lcmaps_log(LOG_ERR, "lcmaps_plugin_tracking_groupid-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n", lcmaps_mode);
        goto fail_tracking_groupid;
    }
    lcmaps_log_debug(5, "%s: starting run\n", __func__);

    /* Get the Secondary Group IDs of the process */
    if (tracking_groupid_get_gidlist(&n_sec_groups_list, &sec_groups_list))
        goto fail_tracking_groupid;

    for (i = 0; i < n_sec_groups_list; i++) {
        /* Is the Secondary GID in the Tracking Group ID range */
        if ((sec_groups_list[i] >= set_groupid_min) &&
            (sec_groups_list[i] <= set_groupid_max)) {

            /* Adding Tracking Group ID */
            addCredentialData(SEC_GID, &(sec_groups_list[i]));
        }
    }

    /* succes */
    if (sec_groups_list) free(sec_groups_list);
    lcmaps_log(LOG_INFO,"%s: tracking groupid plugin succeeded\n", logstr);
    return LCMAPS_MOD_SUCCESS;

fail_tracking_groupid:
    if (sec_groups_list) free(sec_groups_list);
    lcmaps_log(LOG_INFO,"%s: tracking groupid plugin failed\n", logstr);
    return LCMAPS_MOD_FAIL;
}

/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
    char * logstr = "lcmaps_plugin_tracking_groupid-plugin_introspect()";

    lcmaps_log_debug(5,"%s: terminating\n", logstr);

    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   tracking_groupid_get_gidlist
Description:
    Terminate plugin
Parameters:

Returns:
******************************************************************************/
int
tracking_groupid_get_gidlist(int * ngroups, gid_t ** list){
    int t = 0;
    struct group  * g = NULL;

    if (!list || !ngroups)
        return 1; /* Don't even go to failure! */

    /* Get the number of secondary group IDs for root */
    if ((*ngroups = getgroups(0, NULL)) < 0)
        goto failure;

    /* Allocate the list - must be free'd by the function user */
    if ((*list = (gid_t *) malloc(*ngroups * sizeof(gid_t))) == NULL)
        goto failure;

    /* Get the secondary group IDs for root */
    if (getgroups(*ngroups, *list) < 0)
        goto failure;

    /* Returning all Secondary GIDs */
    return 0;

failure:
    if (*list) free(*list);
    return 1;
}

